// LICENSE
// =======
//
// Copyright (c) 2009, Mohammad Abdulfatah (a.k.a. ShaderOp)
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
//	* 	Redistributions of source code must retain the above copyright notice,
//		this list of conditions and the following disclaimer.
//  * 	Redistributions in binary form must reproduce the above copyright notice,
//		this list of conditions and the following disclaimer in the documentation
//		and/or other materials provided with the distribution.
//		
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.
//
//
// DESCRIPTION & USAGE
// ===================
// Visit http://shaderop.com/softimage/geometry-constraint/ for more information.
//
// CONTACT
// =======
// Please feel free to contact me if you have any comments or requests. I'm ShaderOp on
// both CGTalk and Si-Community.com.
//
//
// VERSION HISTORY
// ===============
// v1.2 (10/2/2009)
// 	*	Based on popular demand, Added BlendedGeometryConstraint, in which the target
//		is constrained between a source object and a the surface of polygon mesh. Position
//		on the surface mesh is the position of the source had it been constrained to the mesh
//		using the Geometry constraint. I think I need a technical writer to author a human-
//		readable description of this instead of the Klingon I have just written.
//
// v1.1 (9/30/2009)
//  *	Added on/off switch through the "Active" boolean parameter.
//
// v1.0 (9/29/2009)
//  *	Initial release.
//

function XSILoadPlugin( in_reg )
{
	in_reg.Author = "ShaderOp";
	in_reg.Name = "SO_GeometryConstraintPlugin";
	in_reg.Email = "";
	in_reg.URL = "http://shaderop.com/softimage/geometry-constraint/";
	in_reg.Major = 1;
	in_reg.Minor = 2;

	in_reg.RegisterOperator("SO_GeometryConstraint");
	in_reg.RegisterCommand("SO_ApplyGeometryConstraint", "SO_ApplyGeometryConstraint");
	in_reg.RegisterMenu(siMenuMCPConstrainID, "SO_ApplyGeometryConstraint_Menu", false,false);
	
	in_reg.RegisterOperator("SO_BlendedGeometryConstraint");	
	in_reg.RegisterCommand("SO_ApplyBlendedGeometryConstraint", "SO_ApplyBlendedGeometryConstraint");
	in_reg.RegisterMenu(siMenuMCPConstrainID, "SO_ApplyBlendedGeometryConstraint_Menu", false,false);
	
	//RegistrationInsertionPoint - do not remove this line

	return true;
}
                               
function XSIUnloadPlugin( in_reg )
{
	var strPluginName;
	strPluginName = in_reg.Name;
	return true;
}

function SO_ApplyGeometryConstraint_Init( in_ctxt )
{
	var oCmd;
	oCmd = in_ctxt.Source;
	oCmd.Description = "Constraints an object to the surface of a polymesh.";
	oCmd.SetFlag(siNoLogging,false);

	// TODO: You may want to add some arguments to this command so that the operator
	// can be applied to objects without depending on their specific names.
	// Tip: the Collection ArgumentHandler is very useful
	return true;
}

function SO_ApplyGeometryConstraint_Execute(  )
{		
	if (Selection.Count != 1)
	{
		LogMessage("Please select the object to be constrained before invoking this command.", siError);
		return false;
	}
	
	var target = Selection(0);
	if (target.IsClassOf(siX3DObjectID) == false)
	{
		LogMessage("Invalid selection.", siError);
		return false;
	}
	
	var rtn = PickElement(siObjectFilter, "Polymesh", "Polymesh");
	if (rtn.Value("ButtonPressed") == 0)
	{
		LogMessage("Command canceled.", siWarning);
		return false;
	}
	
	var source = rtn.Value("PickedElement");

	if (source.Type != "polymsh")
	{
		LogMessage("Constraining object should be a polygon mesh. Command aborted.", siError);
		return false;
	}
		
	var newOp = XSIFactory.CreateObject("SO_GeometryConstraint");

	newOp.AddOutputPort(target.FullName + ".kine.global");
	newOp.AddInputPort(target.FullName + ".kine.local");
	newOp.AddInputPort(target.FullName + ".kine.global");
	newOp.AddInputPort(source.FullName + ".kine.local");
	newOp.AddInputPort(source.FullName + ".kine.global");
	newOp.AddInputPort(source.FullName + ".polymsh");
	newOp.Connect();

	newOp.Connect();	
	InspectObj(newOp.FullName);
	
	return newOp;
}

function SO_ApplyGeometryConstraint_Menu_Init( in_ctxt )
{
	var oMenu;
	oMenu = in_ctxt.Source;
	oMenu.Filter = siObjectFilter;
	oMenu.AddCommandItem("To polygon mesh surface","SO_ApplyGeometryConstraint");
	return true;
}

function SO_ApplyBlendedGeometryConstraint_Init( in_ctxt )
{
	var oCmd;
	oCmd = in_ctxt.Source;
	oCmd.Description = "Constraints an object to the surface of a polymesh and an object.";
	oCmd.SetFlag(siNoLogging,false);

	// TODO: You may want to add some arguments to this command so that the operator
	// can be applied to objects without depending on their specific names.
	// Tip: the Collection ArgumentHandler is very useful
	return true;
}

function SO_ApplyBlendedGeometryConstraint_Execute(  )
{		
	if (Selection.Count != 1)
	{
		LogMessage("Please select the object to be constrained before invoking this command.", siError);
		return false;
	}
	
	var target = Selection(0);
	if (target.IsClassOf(siX3DObjectID) == false)
	{
		LogMessage("Invalid selection.", siError);
		return false;
	}
	
	var polymesh = null;
	var obj = null;	
	
	while(polymesh == null || obj == null)
	{
		var rtn = PickElement(siObjectFilter, "Constraining polygon mesh", "Constraining object");
		
		var element = rtn.Value( "PickedElement" );
		var button = rtn.Value( "ButtonPressed" );
		
		if (button == 0)
		{
			LogMessage("Scripted operator creation canceled.", siWarning);
			return null;
		}
		
		if (button == 1)
		{
			if (element.Type != "polymsh")
			{
				LogMessage("You have selected an invalid constraining mesh.", siError);
				polymesh = null;
			}
			else
			{
				polymesh = element;
			}
		}
		
		if (button == 2)
		{
			if (element.IsClassOf(siX3DObjectID) == false)
			{
				LogMessage("You have selected an invalid constraining object.", siError);
				obj = null;
			}
			else
			{
				obj = element;				
			}			
		}	
	}

		
	var newOp = XSIFactory.CreateObject("SO_BlendedGeometryConstraint");

	newOp.AddOutputPort(target.FullName + ".kine.global");
	newOp.AddInputPort(obj.FullName + ".kine.local");
	newOp.AddInputPort(obj.FullName + ".kine.global");
	newOp.AddInputPort(polymesh.FullName + ".kine.local");
	newOp.AddInputPort(polymesh.FullName + ".kine.global");
	newOp.AddInputPort(polymesh.FullName + ".polymsh");
	newOp.Connect();	
	InspectObj(newOp.FullName);
	
	return newOp;
}

function SO_ApplyBlendedGeometryConstraint_Menu_Init( in_ctxt )
{
	var oMenu;
	oMenu = in_ctxt.Source;
	oMenu.Filter = siObjectFilter;
	oMenu.AddCommandItem("Between polygon mesh surface and object","SO_ApplyBlendedGeometryConstraint");
	return true;
}

function SO_GeometryConstraint_Define( in_ctxt )
{
	var oCustomOperator;
	var oPDef;
	oCustomOperator = in_ctxt.Source;
	
	oPDef = XSIFactory.CreateParamDef("active",siBool,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",true,null,null,null,null);	
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("Orient_To_Normal",siBool,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",true,null,null,null,null);	
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("affectedAxisX",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",1.0);	
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("affectedAxisY",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",0.0);
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("affectedAxisZ",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",0.0);
	oCustomOperator.AddParameter(oPDef);

	oCustomOperator.AlwaysEvaluate = false;
	oCustomOperator.Debug = 0;
	return true;
}

function SO_GeometryConstraint_DefineLayout( in_ctxt )
{
	var oLayout,oItem;
	oLayout = in_ctxt.Source;
	oLayout.Clear();
	
	var textbox = null;

  	oLayout.AddItem("active", "Active");
  
	oLayout.AddItem("Orient_To_Normal", "Align to surface normal");
	
	oLayout.AddGroup("Affected Axis");
	
	oLayout.AddRow();
	textbox = oLayout.AddItem("affectedAxisX", "X", siControlString);
	textbox.WidthPercentage = 80;
	textbox.LabelPercentage = 80;
	textbox.LabelMinPixels = 150;	
	oLayout.AddButton("SetXpve", "+X").WidthPercentage = 10;
	oLayout.AddButton("SetXnve", "-X").WidthPercentage = 10;
	oLayout.EndRow();

	oLayout.AddRow();
	textbox = oLayout.AddItem("affectedAxisY", "Y", siControlString);
	textbox.WidthPercentage = 80;
	textbox.LabelPercentage = 80;
	textbox.LabelMinPixels = 150;	
	oLayout.AddButton("SetYpve", "+Y").WidthPercentage = 10;
	oLayout.AddButton("SetYnve", "-Y").WidthPercentage = 10;
	oLayout.EndRow();

	oLayout.AddRow();
	textbox = oLayout.AddItem("affectedAxisZ", "Z", siControlString);
	textbox.WidthPercentage = 80;
	textbox.LabelPercentage = 80;
	textbox.LabelMinPixels = 150;	
	oLayout.AddButton("SetZpve", "+Z").WidthPercentage = 10;
	oLayout.AddButton("SetZnve", "-Z").WidthPercentage = 10;
	oLayout.EndRow();
	
	oLayout.EndGroup();

	return true;
}

function SO_GeometryConstraint_Init( in_ctxt )
{
	return true;
}

function SO_GeometryConstraint_Update( in_ctxt )
{  
    var isActive = in_ctxt.GetParameterValue("active");
    
    if (!isActive)
      return;

	var userData = in_ctxt.UserData;

	if (userData == null)
	{    
		userData = new Object();
		userData.oldPos = XSIMath.CreateVector3();
		userData.oldPointLocatorData = null;

		// Avoiding recreating the object used below helps performance		
		userData.constrainedObjectGlobalPos = XSIMath.CreateVector3();
		userData.constrainedObjectClosestSurfaceLocalPos = XSIMath.CreateVector3();
		userData.closestNormalVector = XSIMath.CreateVector3();
		userData.upVector = XSIMath.CreateVector3();
		userData.xform = XSIMath.CreateTransform();
		userData.rotation = XSIMath.CreateQuaternion();
		userData.originalRotationMatrix = XSIMath.CreateMatrix3();
		userData.qMatrix = XSIMath.CreateMatrix3();
		in_ctxt.UserData = userData;
	}
	
	var constrainedObjectGlobalPos 
		= userData.constrainedObjectGlobalPos;
	var constrainedObjectClosestSurfaceLocalPos	
		= userData.constrainedObjectClosestSurfaceLocalPos;
	var closestNormalVector
		= userData.closestNormalVector;
	var upVector 
		= userData.upVector;		
	var xform 
		= userData.xform;
	var rotation 
		= userData.rotation;
	var originalRotationMatrix 
		= userData.originalRotationMatrix;
	var qMatrix 
		= userData.qMatrix;
	
	var output 					= in_ctxt.OutputTarget;	
	var targetLocalKine			= in_ctxt.GetInputValue(0);
	var targetGlobalKine		= in_ctxt.GetInputValue(1);
	var meshLocalKine			= in_ctxt.GetInputValue(2);
	var meshGlobalKine			= in_ctxt.GetInputValue(3);
	var meshPrimitive			= in_ctxt.GetInputValue(4);
	  
	var Orient_To_Normal		= in_ctxt.GetParameterValue("Orient_To_Normal");
	var affectedAxisX			= in_ctxt.GetParameterValue("affectedAxisX");
	var affectedAxisY			= in_ctxt.GetParameterValue("affectedAxisY");
	var affectedAxisZ			= in_ctxt.GetParameterValue("affectedAxisZ");
	
	upVector.Set(affectedAxisX, affectedAxisY, affectedAxisZ);
	
	targetGlobalKine.Transform.GetTranslation(constrainedObjectGlobalPos);

	var pointLocatorData = userData.oldPointLocatorData;	
	
	
	// Only update the locator data is the constrained object moves.
	// Otherwise, the constrained object will appear to skid along
	// the constraining object's surface if the constraining object
	// is moved.
	if (!constrainedObjectGlobalPos.Equals(userData.oldPos) || pointLocatorData == null)
	{
		var constrainedObjectLocalPos = XSIMath.MapWorldPositionToObjectSpace
		( meshLocalKine.Transform
		, constrainedObjectGlobalPos
		);
	
		pointLocatorData = meshPrimitive.Geometry.GetClosestLocations(constrainedObjectLocalPos.Get2());
	}
	
	var closestPosArray 
		= meshPrimitive.Geometry.EvaluatePositions(pointLocatorData).toArray();

	constrainedObjectClosestSurfaceLocalPos.Set
		( closestPosArray[0]
		, closestPosArray[1]
		, closestPosArray[2]
		);

	
	var constrainedObjectClosestSurfaceGlobalPos = XSIMath.MapObjectPositionToWorldSpace
		( meshLocalKine.Transform
		, constrainedObjectClosestSurfaceLocalPos
		);
		
	userData.oldPos.Copy(constrainedObjectClosestSurfaceGlobalPos);
	
	xform.Copy(targetGlobalKine.Transform);	
	xform.SetTranslation(constrainedObjectClosestSurfaceGlobalPos);			
	
	if (Orient_To_Normal == true && upVector.Length() > 0.00001)
	{
		var closestNormalArray = meshPrimitive.Geometry.EvaluateNormals
				( pointLocatorData
				, 2
				).toArray();
		
		closestNormalVector.Set
			( closestNormalArray[0]
			, closestNormalArray[1]
			, closestNormalArray[2]
			);
		
		closestNormalVector = XSIMath.MapObjectPositionToWorldSpace
			( meshLocalKine.Transform
			, closestNormalVector);

		closestNormalVector.SubInPlace(meshGlobalKine.Transform.Translation);
		closestNormalVector.AddInPlace(targetGlobalKine.Transform.Translation);
		
		closestNormalVector = XSIMath.MapWorldPositionToObjectSpace
			( targetLocalKine.Transform
			, closestNormalVector);		
		
		
		// Need to rotate the constrained object's upVector to match the normal vector.
		// Refer to Matrix and Quaternion FAQ, Q39: How do I generate a rotation matrix to map one
		// vector onto another:
		// http://web.archive.org/web/20041029003853/http:/www.j3d.org/matrix_faq/matrfaq_latest.html#Q39		
		var rotationAxis = XSIMath.CreateVector3();
		rotationAxis.Cross(upVector, closestNormalVector);
		rotationAxis.NormalizeInPlace();
		var rotationAngle = closestNormalVector.Angle(upVector);

		var sinA = Math.sin(rotationAngle/2);
		var qw = Math.cos(rotationAngle/2);
		var qx = rotationAxis.X * sinA;
		var qy = rotationAxis.Y * sinA;
		var qz = rotationAxis.Z * sinA;

		rotation.Set(qw, qx, qy, qz);
		rotation.Normalize();
	
		targetLocalKine.Transform.GetRotationMatrix3(originalRotationMatrix);		
		qMatrix.SetFromQuaternion(rotation);
		qMatrix.MulInPlace(originalRotationMatrix)
		
		xform.SetRotationFromMatrix3(qMatrix);		
	}
	
	// What's below is something of an oddity. The scripted operator
	// needs to be applied to the target's global kine state in order
	// for key animation to work, but working with the local kine state
	// is by far easier, especially when working out the alignment to 
	// the surface normal. The varialble 'xform' is a copy of the target's
	// local kine state, yet assigning it to the global kine state
	// produces the desired results without needing to transform xform
	// to global coordinates. It's probably just my own misunderstanding
	// of how kinematics states actually work, and since it works perfectly
	// I decided not to dwell on it.
	output.Transform = xform;

	userData.oldPointLocatorData = pointLocatorData;
	return true;
}

function SO_GeometryConstraint_SetXnve_OnClicked()
{
	PPG.affectedAxisX.Value = -1;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = 0;	
}

function SO_GeometryConstraint_SetXpve_OnClicked()
{
	PPG.affectedAxisX.Value = 1;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = 0;	
}

function SO_GeometryConstraint_SetYnve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = -1;
	PPG.affectedAxisZ.Value = 0;
}

function SO_GeometryConstraint_SetYpve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = 1;
	PPG.affectedAxisZ.Value = 0;
}

function SO_GeometryConstraint_SetZnve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = -1;
}

function SO_GeometryConstraint_SetZpve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = 1;
}

function SO_GeometryConstraint_Orient_To_Normal_OnChanged()
{
	PPG.affectedAxisX.ReadOnly = !PPG.Orient_To_Normal.Value;
	PPG.affectedAxisY.ReadOnly = !PPG.Orient_To_Normal.Value;
	PPG.affectedAxisZ.ReadOnly = !PPG.Orient_To_Normal.Value;

	PPG.PPGLayout.Item("SetXnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetXpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetYnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetYpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetZnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetZpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	
	PPG.Refresh();
}

function SO_GeometryConstraint_OnInit()
{
	PPG.PPGLayout.Item("SetXnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetXpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetYnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetYpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetZnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetZpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
}


function SO_BlendedGeometryConstraint_Define( in_ctxt )
{
	var oCustomOperator;
	var oPDef;
	oCustomOperator = in_ctxt.Source;

	oPDef = XSIFactory.CreateParamDef("blendWeight",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","", 1.0, 0.0, 1.0, 0.0, 1.0);	
	oCustomOperator.AddParameter(oPDef);
	
	oPDef = XSIFactory.CreateParamDef("active",siBool,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",true,null,null,null,null);	
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("Orient_To_Normal",siBool,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",true,null,null,null,null);	
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("affectedAxisX",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",1.0);	
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("affectedAxisY",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",0.0);
	oCustomOperator.AddParameter(oPDef);

	oPDef = XSIFactory.CreateParamDef("affectedAxisZ",siFloat,siClassifUnknown,siPersistable | siKeyable | siAnimatable,"","",0.0);
	oCustomOperator.AddParameter(oPDef);

	oCustomOperator.AlwaysEvaluate = false;
	oCustomOperator.Debug = 0;
	return true;
}

function SO_BlendedGeometryConstraint_DefineLayout( in_ctxt )
{
	var oLayout,oItem;
	oLayout = in_ctxt.Source;
	oLayout.Clear();
	
	var textbox = null;

  	oLayout.AddItem("active", "Active");

	oLayout.AddItem("blendWeight", "Blend Weight");
	  
	oLayout.AddItem("Orient_To_Normal", "Align to surface normal");
	
	oLayout.AddGroup("Affected Axis");
	
	oLayout.AddRow();
	textbox = oLayout.AddItem("affectedAxisX", "X", siControlString);
	textbox.WidthPercentage = 80;
	textbox.LabelPercentage = 80;
	textbox.LabelMinPixels = 150;	
	oLayout.AddButton("SetXpve", "+X").WidthPercentage = 10;
	oLayout.AddButton("SetXnve", "-X").WidthPercentage = 10;
	oLayout.EndRow();

	oLayout.AddRow();
	textbox = oLayout.AddItem("affectedAxisY", "Y", siControlString);
	textbox.WidthPercentage = 80;
	textbox.LabelPercentage = 80;
	textbox.LabelMinPixels = 150;	
	oLayout.AddButton("SetYpve", "+Y").WidthPercentage = 10;
	oLayout.AddButton("SetYnve", "-Y").WidthPercentage = 10;
	oLayout.EndRow();

	oLayout.AddRow();
	textbox = oLayout.AddItem("affectedAxisZ", "Z", siControlString);
	textbox.WidthPercentage = 80;
	textbox.LabelPercentage = 80;
	textbox.LabelMinPixels = 150;	
	oLayout.AddButton("SetZpve", "+Z").WidthPercentage = 10;
	oLayout.AddButton("SetZnve", "-Z").WidthPercentage = 10;
	oLayout.EndRow();
	
	oLayout.EndGroup();

	return true;
}

function SO_BlendedGeometryConstraint_Init( in_ctxt )
{
	return true;
}

function SO_BlendedGeometryConstraint_Update( in_ctxt )
{  
    var isActive = in_ctxt.GetParameterValue("active");
    
    if (!isActive)
      return;

	var userData = in_ctxt.UserData;

	if (userData == null)
	{    
		userData = new Object();
		userData.oldPos = XSIMath.CreateVector3();
		userData.oldPointLocatorData = null;

		// Avoiding recreating the object used below helps performance		
		userData.constrainedObjectGlobalPos = XSIMath.CreateVector3();
		userData.constrainedObjectClosestSurfaceLocalPos = XSIMath.CreateVector3();
		userData.closestNormalVector = XSIMath.CreateVector3();
		userData.upVector = XSIMath.CreateVector3();
		userData.xform = XSIMath.CreateTransform();
		userData.rotation = XSIMath.CreateQuaternion();
		userData.originalRotationMatrix = XSIMath.CreateMatrix3();
		userData.qMatrix = XSIMath.CreateMatrix3();
		userData.blendedLocation = XSIMath.CreateVector3();

		in_ctxt.UserData = userData;
	}
	
	var constrainedObjectGlobalPos 
		= userData.constrainedObjectGlobalPos;
	var constrainedObjectClosestSurfaceLocalPos	
		= userData.constrainedObjectClosestSurfaceLocalPos;
	var closestNormalVector
		= userData.closestNormalVector;
	var upVector 
		= userData.upVector;		
	var xform 
		= userData.xform;
	var rotation 
		= userData.rotation;
	var originalRotationMatrix 
		= userData.originalRotationMatrix;
	var qMatrix 
		= userData.qMatrix;
	var blendedLocation
		= userData.blendedLocation;
		
	var output 					= in_ctxt.OutputTarget;	
	var targetLocalKine			= in_ctxt.GetInputValue(0);
	var targetGlobalKine		= in_ctxt.GetInputValue(1);
	var meshLocalKine			= in_ctxt.GetInputValue(2);
	var meshGlobalKine			= in_ctxt.GetInputValue(3);
	var meshPrimitive			= in_ctxt.GetInputValue(4);
	  
	var blendWeight				= in_ctxt.GetParameterValue("blendWeight");
	var Orient_To_Normal		= in_ctxt.GetParameterValue("Orient_To_Normal");
	var affectedAxisX			= in_ctxt.GetParameterValue("affectedAxisX");
	var affectedAxisY			= in_ctxt.GetParameterValue("affectedAxisY");
	var affectedAxisZ			= in_ctxt.GetParameterValue("affectedAxisZ");
	
	upVector.Set(affectedAxisX, affectedAxisY, affectedAxisZ);
	
	targetGlobalKine.Transform.GetTranslation(constrainedObjectGlobalPos);

	var pointLocatorData = userData.oldPointLocatorData;	
	
	var constrainedObjectLocalPos = XSIMath.MapWorldPositionToObjectSpace
		( meshLocalKine.Transform
		, constrainedObjectGlobalPos
		);
	
	pointLocatorData = meshPrimitive.Geometry.GetClosestLocations(constrainedObjectLocalPos.Get2());
	
	var closestPosArray 
		= meshPrimitive.Geometry.EvaluatePositions(pointLocatorData).toArray();

	constrainedObjectClosestSurfaceLocalPos.Set
		( closestPosArray[0]
		, closestPosArray[1]
		, closestPosArray[2]
		);
	
	var constrainedObjectClosestSurfaceGlobalPos = XSIMath.MapObjectPositionToWorldSpace
		( meshLocalKine.Transform
		, constrainedObjectClosestSurfaceLocalPos
		);
		
	userData.oldPos.Copy(constrainedObjectClosestSurfaceGlobalPos);
	
	xform.Copy(targetGlobalKine.Transform);
	
	blendedLocation.Copy(constrainedObjectClosestSurfaceGlobalPos);
	blendedLocation.SubInPlace(targetGlobalKine.Transform.Translation);
	blendedLocation.ScaleInPlace(blendWeight);
	blendedLocation.AddInPlace(targetGlobalKine.Transform.Translation);
	
	xform.SetTranslation(blendedLocation);			
	
	if (Orient_To_Normal == true && upVector.Length() > 0.00001)
	{
		var closestNormalArray = meshPrimitive.Geometry.EvaluateNormals
				( pointLocatorData
				, 2
				).toArray();
		
		closestNormalVector.Set
			( closestNormalArray[0]
			, closestNormalArray[1]
			, closestNormalArray[2]
			);
		
		closestNormalVector = XSIMath.MapObjectPositionToWorldSpace
			( meshLocalKine.Transform
			, closestNormalVector);

		closestNormalVector.SubInPlace(meshGlobalKine.Transform.Translation);
		closestNormalVector.AddInPlace(targetGlobalKine.Transform.Translation);
		
		closestNormalVector = XSIMath.MapWorldPositionToObjectSpace
			( targetLocalKine.Transform
			, closestNormalVector);		
			
		var rotationAxis = XSIMath.CreateVector3();
		rotationAxis.Cross(upVector, closestNormalVector);
		rotationAxis.NormalizeInPlace();
		var rotationAngle = closestNormalVector.Angle(upVector) * blendWeight;

		var sinA = Math.sin(rotationAngle/2);
		var qw = Math.cos(rotationAngle/2);
		var qx = rotationAxis.X * sinA;
		var qy = rotationAxis.Y * sinA;
		var qz = rotationAxis.Z * sinA;

		rotation.Set(qw, qx, qy, qz);
		rotation.Normalize();
	
		targetLocalKine.Transform.GetRotationMatrix3(originalRotationMatrix);		
		qMatrix.SetFromQuaternion(rotation);
		qMatrix.MulInPlace(originalRotationMatrix)
		
		xform.SetRotationFromMatrix3(qMatrix);		
	}
	
	output.Transform = xform;

	userData.oldPointLocatorData = pointLocatorData;
	return true;
}

function SO_BlendedGeometryConstraint_SetXnve_OnClicked()
{
	PPG.affectedAxisX.Value = -1;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = 0;	
}

function SO_BlendedGeometryConstraint_SetXpve_OnClicked()
{
	PPG.affectedAxisX.Value = 1;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = 0;	
}

function SO_BlendedGeometryConstraint_SetYnve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = -1;
	PPG.affectedAxisZ.Value = 0;
}

function SO_BlendedGeometryConstraint_SetYpve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = 1;
	PPG.affectedAxisZ.Value = 0;
}

function SO_BlendedGeometryConstraint_SetZnve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = -1;
}

function SO_BlendedGeometryConstraint_SetZpve_OnClicked()
{
	PPG.affectedAxisX.Value = 0;
	PPG.affectedAxisY.Value = 0;
	PPG.affectedAxisZ.Value = 1;
}

function SO_BlendedGeometryConstraint_Orient_To_Normal_OnChanged()
{
	PPG.affectedAxisX.ReadOnly = !PPG.Orient_To_Normal.Value;
	PPG.affectedAxisY.ReadOnly = !PPG.Orient_To_Normal.Value;
	PPG.affectedAxisZ.ReadOnly = !PPG.Orient_To_Normal.Value;

	PPG.PPGLayout.Item("SetXnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetXpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetYnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetYpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetZnve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	PPG.PPGLayout.Item("SetZpve").SetAttribute(siUIButtonDisable, !PPG.Orient_To_Normal.Value);
	
	PPG.Refresh();
}
